import os
from itertools import chain
from typing import Any, Callable, Dict, List, Optional, Tuple, Union

import snowflake.connector

from grai_source_snowflake.models import (
    Column,
    ColumnID,
    Edge,
    EdgeQuery,
    SnowflakeNode,
    Table,
)


def get_from_env(
    label: str,
    default: Optional[Any] = None,
    required: bool = True,
    validator: Optional[Callable] = None,
):
    env_key = f"GRAI_SNOWFLAKE_{label.upper()}"
    result = os.getenv(env_key, default)
    if result is None and required:
        message = (
            f"Establishing snowflake connection requires a {label}. "
            f"Either pass a value explicitly or set a {env_key} environment value."
        )
        raise Exception(message)
    return result if validator is None else validator(result)


class SnowflakeConnector:
    def __init__(
        self,
        account: Optional[str] = None,
        user: Optional[str] = None,
        password: Optional[str] = None,
        warehouse: Optional[str] = None,
        role: Optional[str] = None,
        database: Optional[str] = None,
        schema: Optional[str] = None,
        namespace: Optional[str] = None,
        **kwargs,
    ):
        self.account = get_from_env("account") if account is None else account
        self.user = get_from_env("user") if user is None else user
        self.password = get_from_env("password") if password is None else password
        self.warehouse = get_from_env("warehouse") if warehouse is None else warehouse
        self.role = get_from_env("role", required=False) if role is None else role
        self.database = (
            get_from_env("database", required=False) if database is None else database
        )
        self.schema = (
            get_from_env("schema", required=False) if schema is None else schema
        )
        self.namespace = (
            get_from_env("namespace", "default") if namespace is None else namespace
        )
        self.additional_conn_kwargs = kwargs
        self._connection: Optional[snowflake.connector.SnowflakeConnection] = None

        self._tables: Optional[List[Table]] = None
        self._foreign_keys: Optional[List[Edge]] = None

    def __enter__(self):
        return self.connect()

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.close()

    @property
    def connection_dict(self) -> Dict[str, str]:
        connection_keys = [
            "account",
            "user",
            "password",
            "warehouse",
            "role",
            "database",
            "schema",
        ]
        return {
            key: value
            for key in connection_keys
            if (value := getattr(self, key)) is not None
        }

    def connect(self) -> "SnowflakeConnector":
        if self._connection is None:
            self._connection = snowflake.connector.connect(
                **self.connection_dict, **self.additional_conn_kwargs
            )
        return self

    @property
    def connection(self) -> snowflake.connector.SnowflakeConnection:
        if self._connection is None:
            raise Exception("Not connected, call `.connect()")
        return self._connection

    def close(self) -> None:
        self.connection.close()
        self._connection = None

    def query_runner(self, query: str, param_dict: Dict = {}) -> List[Dict]:
        dict_cursor = self.connection.cursor(snowflake.connector.DictCursor)
        dict_cursor.execute(query, param_dict)
        return dict_cursor.fetchall()  # type: ignore

    def get_tables(self) -> List[Table]:
        """
        Create and return a list of dictionaries with the
        schemas and names of tables in the database
        connected to by the connection argument.
        """

        if self._tables is not None:
            return self._tables

        query = """
	        SELECT table_schema, table_name, table_type
            FROM information_schema.tables
            WHERE table_schema != 'INFORMATION_SCHEMA'
            ORDER BY table_schema, table_name
        """
        res = (
            {k.lower(): v for k, v in result.items()}
            for result in self.query_runner(query)
        )

        additional_args = {
            "namespace": self.namespace,
            "table_database": self.connection_dict["database"],
        }

        tables = [Table(**result, **additional_args) for result in res]
        for table in tables:
            table.columns = self.get_columns(table)

        self._tables = tables
        return self._tables

    def get_columns(self, table: Table) -> List[Column]:
        """
        Creates and returns a list of dictionaries for the specified
        schema.table in the database connected to.
        """

        query = f"""
            SELECT column_name, data_type, is_nullable, column_default
            FROM information_schema.columns
            WHERE table_schema = '{table.table_schema}'
            AND table_name = '{table.name}'
        """

        res = (
            {k.lower(): v for k, v in result.items()}
            for result in self.query_runner(query)
        )

        addtl_args = {
            "namespace": table.namespace,
            "schema": table.table_schema,
            "table": table.name,
        }
        return [Column(**result, **addtl_args) for result in res]

    def get_foreign_keys(self) -> List[Edge]:
        """This needs to be tested / evaluated
        :param connection:
        :param table:
        :return:
        """
        # This query only returns foreign keys, there is also exported keys and primary keys.
        # Information schema itself doesn't carry the column for foreign keys
        # show IMPORTED KEYS in database '{self.connection_dict['database']}'

        if self._foreign_keys is not None:
            return self._foreign_keys

        query = f"""
        show IMPORTED KEYS
        """
        addtl_args = {
            "namespace": self.namespace,
        }

        imported_keys = self.query_runner(query)

        key_rename_map = {
            "constraint_name": "fk_name",
            "self_database": "pk_database_name",
            "self_schema": "pk_schema_name",
            "self_table": "pk_table_name",
            "self_columns": "pk_column_name",
            "foreign_database": "fk_database_name",
            "foreign_schema": "fk_schema_name",
            "foreign_table": "fk_table_name",
            "foreign_columns": "fk_column_name",
        }
        for item in imported_keys:
            item |= {k: item.pop(v) for k, v in key_rename_map.items()}
            item["constraint_type"] = "f"
            item["definition"] = ""  # Available on another table
            item["self_columns"] = item["self_columns"].split(",")
            item["foreign_columns"] = item["foreign_columns"].split(",")

        res = ({k.lower(): v for k, v in result.items()} for result in imported_keys)
        fks = [EdgeQuery(**fk, **addtl_args).to_edge() for fk in res]

        self._foreign_keys = fks
        return self._foreign_keys

    def get_nodes(self) -> List[SnowflakeNode]:
        tables = self.get_tables()
        return list(chain(tables, *[t.columns for t in tables]))

    def get_edges(self) -> List[Edge]:
        tables = self.get_tables()
        edges = list(chain(*[t.get_edges() for t in tables], self.get_foreign_keys()))
        return edges

    def get_nodes_and_edges(self) -> Tuple[List[SnowflakeNode], List[Edge]]:
        nodes = self.get_nodes()
        edges = self.get_edges()
        return nodes, edges



from typing import Any, Dict, List, Literal, Sequence, Type

import grai_schemas.models as base_schemas
from grai_client.schemas.schema import Schema
from grai_schemas import config as base_config
from grai_schemas.models import DefaultValue, GraiNodeMetadata
from multimethod import multimethod

from grai_source_snowflake.models import ID, Column, Edge, Table
from grai_source_snowflake.package_definitions import config


@multimethod
def build_grai_metadata(current: Any, desired: Any) -> None:
    raise NotImplementedError(
        f"No adapter between {type(current)} and {type(desired)} for value {current}"
    )


@build_grai_metadata.register
def build_grai_metadata_from_column(
    current: Column, version: Literal["v1"] = "v1"
) -> base_schemas.ColumnMetadata:

    default_value = current.default_value
    if current.default_value is not None:
        default_value = DefaultValue(
            has_default_value=True, default_value=current.default_value
        )

    data = {
        "version": version,
        "node_type": "Column",
        "node_attributes": {
            "data_type": current.data_type,
            "default_value": default_value,
            "is_nullable": current.is_nullable,
            "is_primary_key": current.is_pk,
        },
    }

    return base_schemas.ColumnMetadata(**data)


@build_grai_metadata.register
def build_grai_metadata_from_node(
    current: Table, version: Literal["v1"] = "v1"
) -> GraiNodeMetadata:
    data = {"version": version, "node_type": "Table", "node_attributes": {}}

    return base_schemas.TableMetadata(**data)


@build_grai_metadata.register
def build_grai_metadata_from_edge(
    current: Edge, version: Literal["v1"] = "v1"
) -> base_schemas.GraiEdgeMetadata:
    data = {"version": version}
    return base_schemas.GraiEdgeMetadata(**data)


@multimethod
def build_snowflake_metadata(current: Any, desired: Any) -> None:
    raise NotImplementedError(
        f"No adapter between {type(current)} and {type(desired)} for value {current}"
    )


@build_snowflake_metadata.register
def build_metadata_from_column(current: Column, version: Literal["v1"] = "v1") -> Dict:
    data = {
        "table_name": current.table,
        "schema": current.column_schema,
    }

    return data


@build_snowflake_metadata.register
def build_metadata_from_edge(current: Edge, version: Literal["v1"] = "v1") -> Dict:
    data = {
        "definition": current.definition,
        "constraint_type": current.constraint_type.value,
    }
    data |= current.metadata if current.metadata is not None else {}

    return data


@build_snowflake_metadata.register
def build_metadata_from_node(current: Table, version: Literal["v1"] = "v1") -> Dict:
    data = {
        "schema": current.table_schema,
        "table_type": current.table_type.value,
    }
    return data


def build_metadata(obj, version):
    return {
        base_config.metadata_id: build_grai_metadata(obj, version),
        config.metadata_id: build_snowflake_metadata(obj, version),
    }


@multimethod
def adapt_to_client(current: Any, desired: Any):
    raise NotImplementedError(f"No adapter between {type(current)} and {type(desired)}")


@adapt_to_client.register
def adapt_column_to_client(current: Column, version: Literal["v1"] = "v1"):
    spec_dict = {
        "name": current.full_name,
        "namespace": current.namespace,
        "display_name": current.name,
        "data_source": config.integration_name,
        "metadata": build_metadata(current, version),
    }
    return Schema.to_model(spec_dict, version=version, typing_type="Node")


@adapt_to_client.register
def adapt_table_to_client(current: Table, version: Literal["v1"] = "v1"):
    metadata = {
        "node_type": "Table",
        "schema": current.table_schema,
    }
    spec_dict = {
        "name": current.full_name,
        "namespace": current.namespace,
        "display_name": current.name,
        "data_source": config.integration_name,
        "metadata": build_metadata(current, version),
    }
    metadata.update(current.metadata)
    return Schema.to_model(spec_dict, version=version, typing_type="Node")


def make_name(node1: ID, node2: ID) -> str:
    node1_name = f"{node1.namespace}:{node1.full_name}"
    node2_name = f"{node2.namespace}:{node2.full_name}"
    return f"{node1_name} -> {node2_name}"


@adapt_to_client.register
def adapt_edge_to_client(current: Edge, version: Literal["v1"] = "v1"):
    spec_dict = {
        "data_source": config.integration_name,
        "name": make_name(current.source, current.destination),
        "namespace": current.source.namespace,
        "source": {
            "name": current.source.full_name,
            "namespace": current.source.namespace,
        },
        "destination": {
            "name": current.destination.full_name,
            "namespace": current.destination.namespace,
        },
        "metadata": build_metadata(current, version),
    }
    return Schema.to_model(spec_dict, version=version, typing_type="Edge")


@adapt_to_client.register
def adapt_list_to_client(objs: Sequence, version: Literal["v1"]) -> List:
    return [adapt_to_client(item, version) for item in objs]



# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': 'src'}

packages = \
['grai_source_snowflake']

package_data = \
{'': ['*']}

install_requires = \
['PyYAML>=6.0,<7.0',
 'grai-client>=0.1.3,<0.2.0',
 'grai-schemas>=0.1.0,<0.2.0',
 'multimethod>=1.8,<2.0',
 'pydantic>=1.9.1,<2.0.0',
 'snowflake-connector-python>=2.8.3,<3.0.0']

setup_kwargs = {
    'name': 'grai-source-snowflake',
    'version': '0.0.2',
    'description': '',
    'long_description': 'None',
    'author': 'Tony Lewis',
    'author_email': 'tony@grai.io',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'None',
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)



from typing import List, Literal, Optional, Tuple

from grai_client.endpoints.client import BaseClient
from grai_client.schemas.edge import Edge
from grai_client.schemas.node import Node
from grai_client.update import update

from grai_source_snowflake.adapters import adapt_to_client
from grai_source_snowflake.loader import SnowflakeConnector


def get_nodes_and_edges(
    connector: SnowflakeConnector, version: Literal["v1"]
) -> Tuple[List[Node], List[Edge]]:
    if version != "v1":
        raise NotImplementedError(
            f"No available implementation for client version {version}"
        )

    with connector.connect() as conn:
        nodes, edges = conn.get_nodes_and_edges()

    nodes = adapt_to_client(nodes, version)
    edges = adapt_to_client(edges, version)
    return nodes, edges


def update_server(
    client: BaseClient,
    namespace: Optional[str] = None,
    account: Optional[str] = None,
    user: Optional[str] = None,
    password: Optional[str] = None,
    role: Optional[str] = None,
    warehouse: Optional[str] = None,
    database: Optional[str] = None,
    schema: Optional[str] = None,
) -> None:

    conn = SnowflakeConnector(
        account=account,
        user=user,
        password=password,
        role=role,
        warehouse=warehouse,
        database=database,
        schema=schema,
        namespace=namespace,
    )
    nodes, edges = get_nodes_and_edges(conn, client.id)
    update(client, nodes)
    update(client, edges)
